from abc import ABC


class MicroStructModel(ABC):
    """Microstructure model."""

    def __init__(self, name):
        """Initialize the name of the microstructure model."""
        self.name = name

    def __str__(self):
        """Representation of the model."""
        return self.name

    @classmethod
    def find_model(cls, model_name):
        return eval(model_name)


class MicroStructModelException(Exception):
    """Handle exceptions related to microstructure models."""
    pass


class InvalidMicroStructModel(MicroStructModelException):
    """Handle exceptions related to wrong microstructure models."""
    pass



from src.nexi.models.microstructure_models import MicroStructModel
from src.nexi.models.functions.nexi_functions import nexi_signal_from_vector, nexi_jacobian_from_vector, \
    nexi_hessian_from_vector, nexi_optimized_mse_jacobian
import numpy as np


class Nexi(MicroStructModel):
    """Neurite Exchange Imaging model for diffusion MRI."""

    # Class attributes
    n_params = 4
    param_names = ["t_ex", "Di", "De", "f"]
    classic_limits = np.array([[1, 150], [0.1, 3.5], [0.1, 3.5], [0.1, 0.9]])
    has_rician_mean_correction = False

    def __init__(self, param_lim=classic_limits, invert_tex=False):
        super().__init__(name='NEXI')
        self.param_lim = param_lim
        self.invert_tex = invert_tex
        self.constraints = [self.constr_on_diffusivities]

    @staticmethod
    def constr_on_diffusivities(param):
        if np.array(param).ndim == 1:
            # Put Di>De
            if param[1] < param[2]:
                # exchange them
                param[1], param[2] = param[2], param[1]
        elif np.array(param).ndim == 2:
            # Put Di>De
            for iset in range(len(param)):
                # if Di < De of the set iset
                if param[iset, 1] < param[iset, 2]:
                    # exchange them
                    param[iset, 1], param[iset, 2] = param[iset, 2], param[iset, 1]
        else:
            raise ValueError('Wrong dimension of parameters')
        return param

    @classmethod
    def get_signal(cls, parameters, acq_parameters):
        """Get signal from single Ground Truth."""
        return nexi_signal_from_vector(parameters, acq_parameters.b, acq_parameters.td)

    @classmethod
    def get_jacobian(cls, parameters, acq_parameters):
        """Get jacobian from single Ground Truth."""
        return nexi_jacobian_from_vector(parameters, acq_parameters.b, acq_parameters.td)

    @classmethod
    def get_hessian(cls, parameters, acq_parameters):
        """Get hessian from single Ground Truth."""
        return nexi_hessian_from_vector(parameters, acq_parameters.b, acq_parameters.td)

    # Optimized methods for Non-linear Least Squares
    @classmethod
    def get_mse_jacobian(cls, parameters, acq_parameters, signal_gt):
        """Get signal from single Ground Truth."""
        return nexi_optimized_mse_jacobian(parameters, acq_parameters.b, acq_parameters.td, signal_gt, acq_parameters.ndim)



# SPDX-FileCopyrightText: 2023-present Quentin Uhl <quentin.uhl@wanadoo.fr>
#
# SPDX-License-Identifier: MIT
__version__ = "1.0.3"
if __name__ == '__main__':
    from .estimate_nexi_noiseless import estimate_nexi_noiseless
    from .estimate_nexi import estimate_nexi
    from .powderaverage.powderaverage import powder_average, save_data_as_npz, normalize_sigma



# Implementation of Non-linear Least Squares
import numpy as np
import scipy.optimize as op
from tqdm import tqdm
from joblib import Parallel, delayed

####################################################################
# Parallelized NLS
####################################################################


def nls_loop(target_signal, microstruct_model, acq_param, nls_param_lim, initial_gt=None):
    """
    Non-linear least squares algorithm for a single ground truth.
    :param target_signal: signal for all b-values (shells) and diffusion times (Deltas)
    :param microstruct_model: microstructure model
    :param acq_param: acquisition parameters
    :param nls_param_lim: bounds for the estimation in the non-linear least squares algorithm
    :param initial_gt: the initial ground truth to start the optimization from. If None, random initializations are used.
    :return: x_sol: estimated ground truth
    :return: x_0: initial ground truth
    """
    # Initial solution
    if not (initial_gt is None):
        x_0 = initial_gt
    else:
        x_0 = np.empty((microstruct_model.n_params,))
        for ind_p in range(microstruct_model.n_params):
            x_0[ind_p] = np.random.uniform(nls_param_lim[ind_p][0], nls_param_lim[ind_p][1])
        for constraint in microstruct_model.constraints:
            x_0 = constraint(x_0)

    # Function to optimize
    optim_fun = lambda x: np.sum(np.square(microstruct_model.get_signal(x, acq_param) - target_signal))

    # Hessian of this function
    # fun_hess = lambda x: optim_fun_hess(x, b, td, Y, acq_dimension)

    # Optimisation function
    if hasattr(microstruct_model, 'get_mse_jacobian'):
        # Jacobian of the Mean Square Error
        fun_jac = lambda x: microstruct_model.get_mse_jacobian(x, acq_param, target_signal)
        result = op.minimize(fun=optim_fun, x0=x_0, method='L-BFGS-B', jac=fun_jac, bounds=nls_param_lim, tol=1e-14)
    else:
        result = op.minimize(fun=optim_fun, x0=x_0, method='L-BFGS-B', bounds=nls_param_lim, tol=1e-14)
    x_sol = result.x
    return [x_sol, x_0]


def nls_parallel(signal, N, microstruct_model, acq_param, nls_param_lim, max_nls_verif=5, initial_gt=None, n_cores=-1):
    """
    Parallelized version of the non-linear least squares algorithm.

    :param signal: signal for all b-values (shells) and diffusion times (Deltas)
    :param N: number of ground truth to estimate
    :param microstruct_model: microstructure model to use
    :param acq_param: acquisition parameters
    :param nls_param_lim: bounds for the estimation in the non-linear least squares algorithm
    :param max_nls_verif: maximum number of times the non-linear least squares algorithm is run for each ground truth
    :param initial_gt: the initial ground truth to start the optimization from. If None, random initializations are used.
    :param n_cores: number of cores to use for the parallelization. If -1, all available cores are used.
    :return: x_sol: estimated ground truth
    :return: x_0: initial ground truth
    """
    x_0 = np.empty([N, microstruct_model.n_params])
    x_sol = np.empty([N, microstruct_model.n_params])
    if initial_gt is None:
        initial_gt = [None for _ in range(N)]
    print('Running Parallelized Non-linear Least Squares')
    if hasattr(microstruct_model, 'get_mse_jacobian'):
        print("Optimization using implemented jacobian.")
    else:
        print("Warning : Jacobian not implemented for this model. Optimization will be slower.")
    x_parallel = Parallel(n_jobs=n_cores)(
        delayed(nls_loop_verified)(signal[irunning], microstruct_model, acq_param, nls_param_lim,
                                   initial_gt[irunning], max_nls_verif) for irunning in tqdm(range(N)))  # n_jobs=-1, verbose=50 for verbose
    # NLS_loop_verified is NLS_loop run several times until NLS doesn't output any boundary. Defined below.
    verif_failed_count = 0
    for index in range(N):
        x_sol[index] = x_parallel[index][0]
        x_0[index] = x_parallel[index][1]
        verif_failed_count += x_parallel[index][2]
    print(f"Failed {max_nls_verif}-times border verification procedure : {verif_failed_count} out of {N}")
    print('Non-linear Least Squares completed')
    return x_sol, x_0

####################################################################
# Verification of each NLS loop max_nls_verif times
####################################################################


def touch_border(x_sol, nls_param_lim, n_param):
    """
    Check if the NLS algorithm has touched the border of the parameter space.
    :param x_sol: estimated ground truth
    :param nls_param_lim: bounds for the estimation in the non-linear least squares algorithm
    :param n_param: number of parameters
    :return: True if the NLS algorithm has touched the border of the parameter space, False otherwise
    """
    for ind_p in range(n_param):
        if x_sol[ind_p] == nls_param_lim[ind_p][0] or x_sol[ind_p] == nls_param_lim[ind_p][1]:
            return True
    return False


def nls_loop_verified(target_signal, microstruct_model, acq_param,
                      nls_param_lim, initial_gt=None, max_nls_verif=5):
    """
    Verification of each NLS loop max_nls_verif times.
    :param target_signal: signal for all b-values (shells) and diffusion times (Deltas)
    :param microstruct_model: microstructure model
    :param acq_param: acquisition parameters
    :param nls_param_lim: bounds for the estimation in the non-linear least squares algorithm
    :param initial_gt: the initial ground truth to start the optimization from. If None, random initializations are used.
    :param max_nls_verif: maximum number of times the non-linear least squares algorithm is run for each ground truth
    :return: x_sol: estimated ground truth
    :return: x_0: initial ground truth
    :return: verif_failed: array of booleans indicating if the NLS algorithm has touched the border of the parameter space
    """
    nls_loop_return = nls_loop(target_signal, microstruct_model, acq_param, nls_param_lim, initial_gt)
    x_sol, x_0 = nls_loop_return[0], nls_loop_return[1]
    iter_nb = 1
    while touch_border(x_sol, nls_param_lim, microstruct_model.n_params) and iter_nb < max_nls_verif:
        nls_loop_return = nls_loop(target_signal, microstruct_model, acq_param, nls_param_lim, initial_gt=None)
        x_sol, x_0 = nls_loop_return[0], nls_loop_return[1]
        iter_nb += 1
    if touch_border(x_sol, nls_param_lim, microstruct_model.n_params) and iter_nb == max_nls_verif:
        verif_failed = 1
        # print("Border touched " + str(int(max_nls_verif)) + " times !")
    else:
        verif_failed = 0
    return [x_sol, x_0, verif_failed]



import numpy as np
from scipy.special import hyp1f1


def l12(x):
    # Confluent hypergeometric function of the first kind, or Kummer's function
    # Another way to define it would be:
    # np.float64(np.exp(x / 2) * ((1-x) * np.float64(scipy.special.iv(0, -x / 2)) - x * np.float64(scipy.special.iv(1, -x / 2))))
    # However, the following version is more stable.
    return hyp1f1(-0.5, 1, x)


def l12_derivate(x):
    # Derivative of the confluent hypergeometric function of the first kind, or Kummer's function
    # z . (d / dz) M(a, b, z) = z.(a-b).M(a, b + 1, z)/b + z.M(a, b, z)
    # z . (d / dz) M(a, b, z) = z.(a/b).M(a + 1, b + 1, z)
    # a = -0.5
    # b = 1
    # (a-b)/b = -1.5
    return -0.5*hyp1f1(0.5, 2, x)


def rice_mean(nu, sigma):
    """
    Mean/Expectation of the Rice distribution

    :param nu: array_like, shape (n_samples, ), mean of the Gaussian distribution
    :param sigma: scalar or array_like with shape (n_samples, ), standard deviation of the Gaussian distribution
    """
    # Scalar case
    if np.isscalar(sigma):
        if sigma == 0:
            return nu
        else:
            x = -0.5 * nu ** 2 / sigma ** 2
            return np.sqrt(np.pi / 2) * sigma * l12(x)
    # Array case
    nan_sigma = np.where(sigma == 0, np.nan, sigma)
    return np.where(sigma == 0, nu, np.sqrt(np.pi / 2) * nan_sigma * l12(-0.5 * nu ** 2 / nan_sigma ** 2))


def rice_mean_and_jacobian(nu, sigma, dnu):
    """
    Mean/Expectation of the Rice distribution and its Jacobian. They are computed together to avoid
    recomputing the confluent hypergeometric function of the first kind, or Kummer's function.

    :param nu: array_like, shape (n_samples, ), mean of the Gaussian distribution
    :param sigma: scalar or array_like with shape (n_samples, ), standard deviation of the Gaussian distribution
    :param dnu: array_like, shape (n_samples, n_parameters), Jacobian of the mean of the Gaussian distribution
    """
    # Scalar case
    if np.isscalar(sigma):
        if sigma == 0:
            dnu_extended = np.hstack((dnu, np.zeros((dnu.shape[0], 1))))
            return nu, dnu_extended
        else:
            x = - nu**2 / (2*sigma**2)
            K = hyp1f1(-0.5, 1, x)
            dK = -0.5*hyp1f1(0.5, 2, x)
            mu = np.sqrt(np.pi / 2) * sigma * K
            dmu_dnu = np.tile((np.sqrt(np.pi / 2) * sigma * dK * (- nu / sigma**2))[:, np.newaxis], dnu.shape[-1]) * dnu
            dmu_dnu = np.hstack((dmu_dnu, np.zeros((dmu_dnu.shape[0], 1))))
            # Uncomment to allow change in sigma (and comment the previous line)
            # dmu_dsigma = np.sqrt(np.pi / 2) * (l12(x) + (nu / sigma)**2 * dK)
            # dmu_dnu = np.hstack((dmu_dnu, np.expand_dims(dmu_dsigma, axis=1)))
            return mu, dmu_dnu
    # Array case
    else:
        nan_sigma = np.where(sigma == 0, np.nan, sigma)
        x = - nu ** 2 / (2 * nan_sigma ** 2)
        K = hyp1f1(-0.5, 1, x)
        dK = -0.5 * hyp1f1(0.5, 2, x)
        mu = np.sqrt(np.pi / 2) * nan_sigma * K
        dmu_dnu = np.tile((np.sqrt(np.pi / 2) * nan_sigma * dK * (- nu / nan_sigma ** 2))[:, np.newaxis], dnu.shape[-1]) * dnu
        dmu_dnu = np.where(np.tile(sigma[:, np.newaxis], dnu.shape[-1]) == 0, dnu, dmu_dnu)
        dmu_dnu = np.hstack((dmu_dnu, np.zeros((dmu_dnu.shape[0], 1))))
        # Uncomment to allow change in sigma (and comment the previous line)
        # dmu_dsigma = np.sqrt(np.pi / 2) * (l12(x) + (nu / sigma)**2 * dK)
        # dmu_dnu = np.hstack((dmu_dnu, np.expand_dims(dmu_dsigma, axis=1)))
        return mu, dmu_dnu


broad5 = lambda matrix: np.tile(matrix[..., np.newaxis], 5)
broad6 = lambda matrix: np.tile(matrix[..., np.newaxis], 6)



import os
import sys
import numpy as np
import nibabel as nib

# Get the absolute path to the root directory of your project
project_root = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
# Append the project root to sys.path
sys.path.append(project_root)

from src.nexi.models.functions.nexi_functions import nexi_signal


def generate_phantom():
    """
    Generate a phantom with 27 voxels
    """
    # Set the seed for the global random number generator
    np.random.seed(42)
    # Create a specific instance of RandomState for reproducibility
    rng = np.random.RandomState(42)
    # Generate random numbers using the rng instance : rng.rand(5)

    # Initialize
    signals = []
    parameters = []
    sigmas = []

    # Parameters
    b = np.array([0, 1, 0, 2, 4, 0, 1, 2, 5])
    t = np.array([20, 20, 30, 30, 30, 40, 40, 40, 40])

    for ind in range(27):
        tex = 2 + 98 * rng.rand()
        Di = 2 + rng.rand()
        De = 0.5 + rng.rand()
        Dp = De
        f = rng.rand()
        # Genreate associated signal
        single_signal = nexi_signal(tex, Di, De, Dp, f, b, t)
        magnitude = 10 + 200 * rng.rand()
        single_signal = magnitude * single_signal
        signals.append(single_signal)
        # Random sigma
        sigma = (0.01 + 0.3 * rng.rand()) * magnitude
        sigmas.append(sigma)
        parameters.append([tex, Di, De, f])
    signals = np.array(signals)
    sigmas = np.array(sigmas)
    parameters = np.array(parameters)

    # Reshape and add problematic values
    image = signals.reshape((3, 3, 3, len(b)))
    image[1, 2, 0, 2] = np.nan
    image[1, 2, 1, 0] = np.nan
    image[2, 1, 0, 1] = np.nan
    lowb_noisemap = sigmas.reshape(3, 3, 3)
    lowb_noisemap[0, 1, 1] = np.nan

    # Save the powder average
    image_nii = nib.Nifti1Image(image, affine=np.eye(4))
    nib.save(image_nii, 'tests/nexi/data/phantom.nii.gz')

    lowb_noisemap_nii = nib.Nifti1Image(lowb_noisemap, affine=np.eye(4))
    nib.save(lowb_noisemap_nii, 'tests/nexi/data/lowb_noisemap.nii.gz')

    # Create a mask
    mask = np.ones((3, 3, 3))
    mask[1, 2, 0] = 0
    mask[1, 1, 1] = 0
    mask_nii = nib.Nifti1Image(mask, affine=np.eye(4))
    nib.save(mask_nii, 'tests/nexi/data/mask.nii.gz')

    # Save diffusion times
    np.savetxt('tests/nexi/data/phantom.bval', b, fmt='%.4f')
    np.savetxt('tests/nexi/data/phantom.td', t, fmt='%.4f')



import numpy as np
from tqdm import tqdm
import itertools
from joblib import Parallel, delayed
from src.nexi.models.functions.rician_mean import rice_mean


def find_nls_initialization(signal, sigma, nb_estimates, acq_param, microstruct_model, grid_search_param_lim,
                            debug=False):
    """
    Find the initialization of the Non-linear least squares algorithm.
    :param signal: signal for all b-values (shells) and diffusion times (Deltas)
    :param sigma: noise standard deviation
    :param nb_estimates: number of ground truth to estimate
    :param acq_param: acquisition parameters
    :param microstruct_model: microstructure model
    :param grid_search_param_lim: bounds for the grid search
    :param debug: if True, print debug information
    :return: initial_gt: array of optimized initial ground truths for the NLS algorithm
    """
    # Access to grid search hyperparameters
    n_cores = -1  # multiprocessing.cpu_count()
    overlap = 1  # from 0 to 1 : 0 mean no overlap of the grid ranges, 1 means complete overlap
    nb_points = [15, 12, 8, 8]

    # Check that the number of parameters in the grid search and in the model are the same
    assert microstruct_model.n_params == len(grid_search_param_lim), "Number of parameters in the grid search and in the model are different"

    # Define number of moving parameters
    if microstruct_model.has_rician_mean_correction:
        n_moving_param = microstruct_model.n_params - 1  # -1 because sigma is fixed
    else:
        n_moving_param = microstruct_model.n_params

    print(f"Initializing first grid search inside :\n{grid_search_param_lim}\nusing model {microstruct_model.name}\n...")

    # Generate the parameter grid combinations
    param_grid = [np.linspace(grid_search_param_lim[p][0], grid_search_param_lim[p][1], nb_points[p] + 2)[1:-1].tolist() for p in range(n_moving_param)]
    grid_combinations = list(itertools.product(*param_grid))
    parameters = np.array(grid_combinations)

    # Generate the non-corrected signal dictionary (or grid)
    if microstruct_model.has_rician_mean_correction:
        non_corrected_model = microstruct_model.non_corrected_model
        print(f"Generating {non_corrected_model.name} signal dictionary")
        signal_dict = np.array(Parallel(n_jobs=n_cores)(delayed(non_corrected_model.get_signal)(params, acq_param) for params
                                                        in tqdm(grid_combinations)))
    else:
        signal_dict = np.array(Parallel(n_jobs=n_cores)(delayed(microstruct_model.get_signal)(params, acq_param) for params
                                                        in tqdm(grid_combinations)))
    if debug:
        print("Signal dictionary shape : ", signal_dict.shape)

    # Find the initial ground truth whether the model has a Rician mean correction or not
    print("Extracting initial Ground Truth from dictionary")
    if microstruct_model.has_rician_mean_correction:
        initial_gt = np.array(Parallel(n_jobs=n_cores)(
            delayed(least_square_argmin_rician_corrected)(signal[i], sigma[i], signal_dict, parameters, acq_param.ndim)
            for i in tqdm(range(nb_estimates))))
        initial_gt = np.hstack((initial_gt, np.array([sigma]).T))
    else:
        initial_gt = np.array(Parallel(n_jobs=n_cores)(delayed(least_square_argmin)(signal[i], signal_dict, parameters, acq_param.ndim) for i in tqdm(range(nb_estimates))))
    if debug:
        print('initial_gt shape : ', initial_gt.shape)

    # Add randomness to the initial GT
    for parameter in range(len(grid_search_param_lim)-1):
        initial_gt[:, parameter] += (1+overlap)*(grid_search_param_lim[parameter][1]-grid_search_param_lim[parameter][0])/(nb_points[parameter]+1)*(np.random.rand(nb_estimates) - 0.5)

    # Print the first initial ground truth to check
    if debug:
        for p_ind, param_name in enumerate(microstruct_model.param_names):
            print(f'First initial_gt {param_name} : ', initial_gt[0, p_ind])
    print("First Grid Search completed\n")
    return initial_gt


def fix_parameter_limits_for_grid_search(param_lim):
    """
    Fix the parameter limits for the grid search.
    :param param_lim: bounds for the estimation in the non-linear least squares algorithm
    :return: param_lim: fixed bounds
    """
    updated_param_lim = np.copy(param_lim)
    epsilon = 1e-3
    for parameter_index in range(4):
        if param_lim[parameter_index][0] == 0:
            updated_param_lim[parameter_index][0] = epsilon
    if param_lim[3][1] == 1:
        updated_param_lim[3][1] = 1-epsilon
    return updated_param_lim


def least_square_argmin(signal_i, signal_dict, parameters, acq_param_ndim):
    """
    Compute the argmin of the least squares function for a given signal.
    :param signal_i: signal for all b-values (shells) and diffusion times (Deltas)
    :param signal_dict: signal dictionary
    :param parameters: parameters of the signal dictionary
    :param acq_param_ndim: number of dimensions of the acquisition parameters
    :return: argmin of the least squares function
    """
    if acq_param_ndim == 2:
        lstsqr = np.sum(np.square(np.expand_dims(signal_i, axis=0) - signal_dict), axis=(-2, -1))
    elif acq_param_ndim == 1:
        lstsqr = np.sum(np.square(np.expand_dims(signal_i, axis=0) - signal_dict), axis=-1)
    else:
        raise NotImplementedError
    return parameters[np.unravel_index(lstsqr.argmin(), lstsqr.shape)]


def least_square_argmin_rician_corrected(signal_i, sigma_i, initial_signal_dict, parameters, acq_param_ndim):
    """
    Compute the argmin of the least squares function for a given not yet corrected signal and noise level.
    :param signal_i: signal for all b-values (shells) and diffusion times (Deltas)
    :param signal_dict: signal dictionary
    :param parameters: parameters of the signal dictionary
    :param acq_param_ndim: number of dimensions of the acquisition parameters
    :return: argmin of the least squares function
    """
    # Update the signal dictionary with the noise level
    corrected_signal_dict = rice_mean(initial_signal_dict, sigma_i)
    # Compute the argmin of the least squares function
    if acq_param_ndim == 2:
        lstsqr = np.sum(np.square(np.expand_dims(signal_i, axis=0) - corrected_signal_dict), axis=(-2, -1))
    elif acq_param_ndim == 1:
        lstsqr = np.sum(np.square(np.expand_dims(signal_i, axis=0) - corrected_signal_dict), axis=-1)
    else:
        raise NotImplementedError
    return parameters[np.unravel_index(lstsqr.argmin(), lstsqr.shape)]
